package Setup;

use strict;
use warnings;

=head1 Setup

Set of routines to parse config and setup parameters.
Used by report and data modules to get things from right place
and in right form.
These functions define syntax of parameters, but not semantics.

I want config and setup stuff to be accessed from one place.

I dunno why...

=head2 Remarks

TODO: enhanced selection syntax should be invented to
      support BY-grouping, inclusion functions in groups
      and WHERE filtering.
      Should be something like:
      select: GROUP, FUN by GROUP for WHERE-MATCH
      with defaults to 'ALL'.
      Dunno what to do with GROUP / GROUP selections.
      Functions could be added to only one group in 
      splited selection.
      They say, there's some SQR language...
      Ain't i inventing a wheel ?

=cut

use vars qw#@EXPORT_OK
		%datafields @idc_stack @ibc_opstack $ibc_lex
		$ibc_chr#;
use Exporter;
use Carp qw#carp cluck croak confess#;
use Config qw#get_config#;
use Utils qw#cmp_arrays normname#;

@Setup::ISA = ('Exporter');

@EXPORT_OK = qw#dataclassen datalabels datatypes selections#;

$Data::VERSION = '0.1.0';
$Data::NAME = 'Setup';

import Parser qw#%datafields#;

if ($::DEBUG & 0x01)
{
	print $/, $Data::NAME, ': Fields in %datafields:';
	for (keys %datafields)
	{
		print $/, $Data::NAME, ':   ', $_, "\t", $datafields{$_};
	}
}

#*datafields = *Parser::datafields;

=head2 datalabels datatypes dataclassen

Scan %Parser::datafields in form ( label => 'class:type' ) and
return requested part of info in form (label => info).

I'm not sure all these data will be stored in one hash
and in this form in future releases.

=cut

sub datalabels
{
    my %labels=();
    return keys %datafields;
}

sub datatypes 
{
    my %types=();
    while( my ($label,$def)= each %datafields ) {
	my $fld = normname($label);
	$def = ($def =~ /^.*:(.+)$/) ? $1 : undef;
	$types{$fld} = $def;
	};
    return %types;
}

sub dataclassen 
{
    my %classes=();
    while( my ($label,$def)= each %datafields ) {
	my $fld = normname($label);
	$def = ($def =~ /^([^:]+).*$/)? $1 : undef;
	$classes{$fld} = $def;
	};
    return %classes;
}

=head2 selections

Parses all 'select' directives.  Recognizes function
definitions and grouping (or single key) selections.
Does not check if everything is valid, just syntax.

For each function selection element is:
  [ FUN $fun $arg ]

select: foo(bar)	    => FUN foo bar
select: foo(bar),foo(baz)   => 2 elements:
                               FUN foo bar
			       FUN foo baz

For each grouping or single key selection:
  [ GRP @grouping ]

Here @grouping is a list of sets of groups defined in directive.
Comma-separated list produces single set.
Slash-separated list produces several sets.

select: foo		    => GRP [(foo)]
select: foo,bar+baz	    => GRP [(foo) (bar baz)]
select: foo / bar,baz	    => GRP [(foo)] [(bar) (baz)]

This sub only defines syntax of 'select' directive, and so
it should not join selections or split groups or whatever.

Such complicated return value is to make report modules independant
on how grouping and splitting are defined, and setup parsing - independant
on how config parameters are going to be used.

For other types of selection, another TAGs should be used,
to do not rewrite all code, then report modules ignoring
unknown tags will work fine.

=cut

#
# experimental:
# FUN $fun $arg \@ongroups \@bygroups
#
sub selections
{
    my @selections;
    my $funmatch=qr/(\w+)\((.*?)\)/;
    return () unless (get_config ('select'));
    foreach my $sel (get_config ('select')) { 
	if( $sel =~ /^$funmatch/ ) {
	    my @funcs = ();
	    my @ongroups = ();
	    my @bygroups = ();
	    while( $sel =~ s/^$funmatch,?\s*// ) {
		my ($fun,$fld)=(uc($1),$2);
		push @funcs, [ $fun, normname($fld) ];
		}
	    while( $sel =~ s/\b(by|on)\b\s*(.*)\s*(?=by|on|$)// ) {
		my ($pro,$expr) = ($1,$2);
		if( $pro =~ /by/ ) { push @bygroups, evalg($expr); }
		if( $pro =~ /on/ ) { push @ongroups, evalg($expr); }
		}
	    if( $sel ) {
		die "\nsetup: ERROR: syntax error in 'select', near '$sel'\n";
		}
	    # exp
	    if( @ongroups or @bygroups ) {
		foreach my $f ( @funcs ) {
		    push @selections, [ 'SEL', @$f, \@ongroups, \@bygroups ];
		    }
		}
	    else {
		foreach my $f ( @funcs ) {
		    push @selections, [ 'FUN', @$f ];
		    }
		}
	    }
	else {
	    # set of groups with ('/') among them
	    my @groups=evalg($sel);
	    # groups, splited by marker ('/')
	    my @selection=();
	    my @subset;
	    while( my $g=shift @groups ) {
		if( cmp_arrays($g,['/']) ) {
		    push @selection, [@subset];
		    @subset=();
		    next;
		    }
		push @subset,$g;
		}
	    push @selection,[@subset];
	    push @selections, [ 'GRP', @selection ];
	    }
	}
    return @selections;
}

# hairy index algebra
# L[...] means map L, <...> - resulting grouping,
# A and B - groupings, a and b - indexes
# (a),(b) = <(a) (b)>
#  A , B  = <A B>
# (a)+(b) = <(a b)>
#  A + B  = A[B[a+b]]
# (a)*(b) = <(a b) (b a)>
#  A * B  = A+B, B+A

use vars qw(@idc_stack @ibc_opstack);

# content of stack and result of operation is
# list of refs to lists of refs to lists of keyfields, ...just that.
# that is - several lists of groupings of keys to combine.
# In the end, the should only one group set left - as usually.
sub index_mux { my($A,$B)=@_; return [map { my $b=$_; map { [@$_,@$b] } @$A } @$B]; }
sub index_mul { my($A,$B)=@_; return [@{index_mux($A,$B)},@{index_mux($B,$A)}]; }

# reverse polish calculator
@idc_stack=();
sub idc 
{
    my $op = shift;
    if( ref($op) eq 'ARRAY' ) { push @idc_stack, $op; return 1; }
    if( scalar @idc_stack < 2 ) { print STDERR "\n$Setup::NAME: ERROR: missing operand for '$op'"; return 0; }
    my $B = pop @idc_stack;
    my $A = pop @idc_stack;
    if(    $op eq '+'  ) { push @idc_stack,index_mux($A,$B); }
    elsif( $op eq '*'  ) { push @idc_stack,index_mul($A,$B); }
    elsif( $op eq ','  ) { push @idc_stack,[@$A,@$B]; }
    else {
	print STDERR "\n$Setup::NAME: ERROR: bogus operator '$op'";
	return 0;
	}
    return 1;
}

# normal infex calculator with two hacks:
#  ~  is back-comma hack: collapses pending expression with
#     priority of comma, pushes affected operand back into
#     idc stack to pend for next expression.
#     Looks like substituting foo~ => foo,foo
#
#  /  is self-inserting operator to help markup splitting:
#     it collapses pending expression (with priority of '='),
#     joins it as single-element index, and then puts comma
#     to opstack to join rest of expr;
#     Looks like substituting "foo / bar" => foo,/,bar
#     but with lower priority;
#
#  =  is just end-of-expr markup: collapses pending expression
#     because of lowest priority.

# ibc_lex is in reverse priority order
# such order is to make invalid operators (got -1 priority)
# to fall into dc and catch an error
# most syntax is ~checked~ when '/' or '=' or ','
# throughs pending operators or parenthesis into idc
# forcing it to swear in polish
use vars qw($ibc_lex $ibc_chr);
$ibc_lex="+**,~)(/=";
$ibc_chr="+*,~/";
@ibc_opstack=();
sub ibc
{
    my $expr=shift;
    my $o;
    while( $expr =~ s/^\s*([()=]|[$ibc_chr]|(\*\*)|[^$ibc_lex\s]+)// ) {
	my $t=$1;
	# dmp($t);
	if( $t eq '(') { push @ibc_opstack,$t; next; }
	if( $t=~/^[$ibc_lex]+$/ ) {
	    my $btop=$idc_stack[$#idc_stack] if( $t eq '~');
	    while( $o = $ibc_opstack[$#ibc_opstack] ) {
		# moved out from while() to catch syntax errors
		last if( index($ibc_lex,$t) <= index($ibc_lex,$o) ); 
		pop @ibc_opstack;
		# print "\n<<$o";
		$o=',' if( $o eq '/' );
		idc($o) or return 0;
		}
	    if( $o and "$o$t" eq "()" ) { pop @ibc_opstack; next; }

	    if( $t eq '~')  { idc($btop); $t=','; } 
	    if( $t eq '/' ) { idc([['/']]); idc(','); }

	    push @ibc_opstack,$t unless($t eq '=');
	    next;
	    }
	# print "\n<<($t)";
	idc([[$t]]) or return 0;
	}
    return 1;
}

sub dmp
{
    print "---------";
    print map { "\n[".
			   join(" ",
				map { "(".join(" ",@$_).")" } @$_
				)."]"
		     } @idc_stack;
    print "\n[",join(" ",@ibc_opstack),"]";
}

=head2 evalg $EXPR

Evaluates grouping expression.
Returns list of indexes.

=cut

sub evalg
{
    my $expr=shift;
    return () unless($expr);
    my @groups=();
    ibc($expr.'=');
    # dont move this check out 
    # because it is realy a syntax error and it
    # cannot be handled by polish calculator.
    if( scalar @idc_stack > 1 or scalar @ibc_opstack ) { 
	dmp();
	die "\n$Setup::NAME: FATAL: syntax error in grouping expression: $expr\n"; 
	}
    @groups=@{pop @idc_stack};
    @groups=map {[map(normname($_),@$_)]} @groups;
    # remove dublicated fields, leaving first occurence
    @groups = map {
		   my @g=@$_;
		   my @ng=();
		   foreach my $f (@g) {
		       push @ng,$f unless( grep($f eq $_,@ng) );
		       };
		   $_=[@ng];
		   } @groups;

    return @groups;
}

1;
