package Parser;

use strict;
use warnings;
use vars qw(%extra %datafields);

use Exporter;
use WebserverTools qw(%month_numbers detect_referer detect_browser detect_os
	extract_data);

@Parser::EXPORT_OK = qw(parse extra %datafields);

@Parser::ISA = ('Exporter');

# FIXME pass date and hour in seconds to properly format and sort.

$extra{'total'} = 0;
$extra{'days'} = {};
#$extra{'codered'} = 0;
#$extra{'nimda'} = 0;
$extra{'search_terms'} = {};

%datafields = (	
    host	=> 'key:host',
    user	=> 'key',
    date	=> 'key',
    hour	=> 'key',
    domain	=> 'key:host',
    tld		=> 'key',
    file	=> 'key',
    status	=> 'key',
    bytes	=> 'amount:bytes',
    requests	=> 'amount:number',
    browser	=> 'key',
    os		=> 'key',
    referer	=> 'key'
);

# This needs to be done at runtime, since Data uses Setup which relies on
# %datafields to be defined  -octo
require Data;
import Data qw#store#;

print "\nparser/ncsa: Using NCSA/ACL (apache combined log) format" if $::DEBUG;

return (1);

sub parse
{
	my $line = shift or return undef;
	if ($line =~ /^(\S+)\s(\S+)\s(.+?)\s\[(.+?)\]\s\"(.+?)\"\s(\d+)\s(.+?)\s\"(.+?)\"\s\"(.+?)\"(?:\s\"(.+?)\")?$/)
	{
# Initialize the variables that we can get out of
# each line first..
		my ($host, $ident, $user, $date, $request, $status,
			$bytes, $referer, $browser, $cookie) =
		($1, $2, $3, $4, $5, $6, $7, $8, $9, $10);

# And now initialize all the variables we will use
# to get more information out of each field..
		my ($day, $month, $year, $hour, $minute, $second) =
			$date =~ m#(\d\d)/(\w{3})/(\d{4}):(\d\d):(\d\d):(\d\d)#;

		my $n_date = $year . '-' . $month_numbers{$month} . '-' . $day;

		my ($method, $file, $params) =
			$request =~ m#(\S+) ([^\?\s]+)\??(\S*)#
			or return (0);
		
#		my $attack = 0;
#		if (lc ($file) eq '/default.ida') { ++$extra{'codered'}; $attack = 1; }
#		elsif (lc ($file) eq '/scripts/root.exe') { ++$extra{'nimda'}; $attack = 1; }
#		elsif ($file =~ /root\.exe|winnt/i) { $attack = 1; }
#		if ($attack and !$::config->{'inc_attacks'}) { return 'attack'; }

		if (($user ne '-') and ($status >= 400) and ($status < 500))
		{
			$user = 'invalid';
		}

		if ($user eq '-') { $user = 'unknown'; }
		if ($bytes eq '-') { $bytes = 0; }

		my ($domain, $tld);
		if ($host =~ m/([^\.]+)\.([a-z]{2,})$/i)				
		{
			$tld = lc ($2);
			$domain = lc ($1) . ".$tld";
		}
		else
		{       
			$domain = 'UNRESOLVED';
			$tld = 'UNRESOLVED';
		}

		my $os = detect_os ($browser);
		my $browser_name = detect_browser ($browser);
		my @search_terms = extract_data ($referer);
		$referer = detect_referer ($referer);
		if ($referer eq '-') { $referer = ''; }

		if (!defined $extra{'days'}->{$n_date}) { $extra{'days'}->{$n_date} = 1; }
		++$extra{'total'};

		if (scalar @search_terms)
		{
			print $/, __FILE__, ": Search Terms: ",
				join (' ', @search_terms)
				if ($::DEBUG & 0x100);
			
			$extra{'search_terms'}->{$_}++ for (@search_terms);
		}

		my %combined = (
					'host'		=>	$host,
					'user'		=>	$user,
					'date'		=>	$n_date,
					'hour'		=>	$hour,
					'browser'	=>	$browser_name,
					'os'		=>	$os,
					'domain'	=>	$domain,
					'tld'		=>	$tld,
					'file'		=>	$file,
					'referer'	=>	$referer,
					'status'	=>	$status,
					'bytes'		=>	$bytes,
					'requests'	=>	1
				);
		store (\%combined);
	}
	elsif ($::DEBUG)
	{
		chomp ($line);
		print STDERR "\nUnable to parse: '$line'";
	}
}

sub extra
{
	my ($average, $days) = (0, 0);
	$days = scalar keys %{$extra{'days'}};
	$average = int (($extra{'total'} * 10) / $days) / 10;
	$::EXTRA->{'Total requests'} = $extra{'total'};
	$::EXTRA->{'Averange requests per day'} = $average;
	$::EXTRA->{'Reporting period'} = "$days days";
#	$::EXTRA->{'CodeRed attacks'} = $extra{'codered'};
#	$::EXTRA->{'Nimda attacks'} = $extra{'nimda'};
	
	my @sorted_terms = sort
		{ $extra{'search_terms'}->{$b} <=> $extra{'search_terms'}->{$a} }
		(keys %{$extra{'search_terms'}});
	
	if (@sorted_terms)
	{
		my $max = $extra{'search_terms'}->{$sorted_terms[0]};
		my @scalar_terms = ();
		
		while (@sorted_terms and
			($extra{'search_terms'}->{$sorted_terms[0]} / $max) > 0.1)
		{
			$_ = shift (@sorted_terms);
			
			push (@scalar_terms,
				sprintf ("%s (%u)",
					$_, $extra{'search_terms'}->{$_})
			);
		}
		$::EXTRA->{'Search terms used'} = join (', ', @scalar_terms);

		if (@sorted_terms)
		{
			my $skipped = scalar (@sorted_terms);
			$::EXTRA->{'Search terms used'} .= " [$skipped more skipped]";
		}
	}
}
